/** RemapFrames
  *     An AviSynth plug-in that remaps the frame indices in a clip as
  *     specified by an input text file or by an input string.
  *
  * Last modified: 2004-12-24
  *
  * Copyright (c) 2004 James D. Lin
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
  * met:
  *
  * * Redistributions of source code must retain the above copyright
  *   notice, this list of conditions and the following disclaimer.
  *
  * * Redistributions in binary form must reproduce the above copyright
  *   notice, this list of conditions and the following disclaimer in the
  *   documentation and/or other materials provided with the distribution.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
  * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
  * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */

#include <cstdio>
#include <cassert>

#include "windows.h"
#include "avisynth.h"

#include "ScopeGuard.h"

#include "RemapFrames.h"
#include "RemapFramesParser.h"



// CLASS DEFINITIONS ---------------------------------------------------

/** initSimpleMode
  *
  *     Initializer for RemapFramesSimple.
  *
  * PARAMETERS:
  *     IN filenameP - the name of the text file containing the frame
  *                      mappings;
  *                    may be NULL
  *     IN mappingsP - string containing additional frame mappings;
  *                    may be NULL
  *     IN/OUT envP  - pointer to the AviSynth scripting environment
  *
  * PRE:
  *     Either filenameP or mappingsP must be NULL, but not both.
  */
void RemapFrames::initSimpleMode(const char* filenameP, const char* mappingsP,
                                 IScriptEnvironment* envP)
{
    if (filenameP == NULL && mappingsP == NULL)
    {
        envP->ThrowError("RemapFramesSimple: no filename nor mappings specified");
    }
    else if (filenameP != NULL && mappingsP != NULL)
    {
        envP->ThrowError("RemapFramesSimple: filename and mappings cannot be used together");
    }

    try
    {
        if (mappingsP != NULL)
        {
            RemapFramesParser parser(mappingsP, &indices, sourceClip->GetVideoInfo().num_frames);
            try
            {
                vi.num_frames = parser.parseSimple();
            }
            catch (RemapFramesParser::MalformedException&)
            {
                envP->ThrowError("RemapFramesSimple: parse error in <mappings> string "
                                 "(line offset %u, column %u)",
                                 parser.getLineNumber(), parser.getColumn());
            }
            catch (RemapFramesParser::OverflowException&)
            {
                envP->ThrowError("RemapFramesSimple: integer overflow in <mappings> string "
                                 "(line offset %u, column %u)",
                                 parser.getLineNumber(), parser.getColumn());
            }
            catch (RemapFramesParser::BadValueException& e)
            {
                envP->ThrowError("RemapFramesSimple: value out of bounds in <mappings> string: %d "
                                 "(line offset %u)",
                                 e.val, parser.getLineNumber());
            }
        }
        else if (filenameP != NULL)
        {
            FILE* fileP = fopen(filenameP, "r");
            ON_BLOCK_EXIT(fclose, fileP);

            if (fileP == NULL)
            {
                envP->ThrowError("RemapFramesSimple: error opening file \"%s\"", filenameP);
            }
            else
            {
                RemapFramesParser parser(fileP, &indices, sourceClip->GetVideoInfo().num_frames);
                try
                {
                    vi.num_frames = parser.parseSimple();
                }
                catch (RemapFramesParser::MalformedException&)
                {
                    envP->ThrowError("RemapFramesSimple: parse error "
                                     "(%s, line %u, column %u)",
                                     filenameP, parser.getLineNumber(), parser.getColumn());
                }
                catch (RemapFramesParser::OverflowException&)
                {
                    envP->ThrowError("RemapFramesSimple: integer overflow "
                                     "(%s, line %u, column %u)",
                                     filenameP, parser.getLineNumber(), parser.getColumn());
                }
                catch (RemapFramesParser::BadValueException& e)
                {
                    envP->ThrowError("RemapFramesSimple: value out of bounds: %d "
                                     "(%s, line %u)",
                                     e.val, filenameP, parser.getLineNumber());
                }
            }
        }
    }
    catch (std::bad_alloc&)
    {
        envP->ThrowError("RemapFramesSimple: insufficient memory");
    }
}


/** initReplaceSimpleMode
  *
  *     Initializer for RemapFrames.
  *
  * PARAMETERS:
  *     IN filenameP - the name of the text file containing the frame
  *                      mappings;
  *                    may be NULL
  *     IN mappingsP - string containing additional frame mappings;
  *                    may be NULL
  *     IN/OUT envP  - pointer to the AviSynth scripting environment
  *
  * PRE:
  *     At least one of filenameP or mappingsP must not be NULL.
  */
void RemapFrames::initReplaceSimpleMode(const char* filenameP, const char* mappingsP,
                                        IScriptEnvironment* envP)
{
    if (filenameP == NULL && mappingsP == NULL)
    {
        envP->ThrowError("ReplaceFramesSimple: no filename nor mappings specified");
    }

    try
    {
        // Initialize the values in indices.
        // Each frame by default gets mapped to itself.
        indices.resize(vi.num_frames);
        for (int i = 0; i < vi.num_frames; i++)
        {
            indices[i].clipIndex = 0;
            indices[i].frame = i;
        }

        if (filenameP != NULL)
        {
            FILE* fileP = fopen(filenameP, "r");
            ON_BLOCK_EXIT(fclose, fileP);

            if (fileP == NULL)
            {
                envP->ThrowError("ReplaceFramesSimple: error opening file \"%s\"", filenameP);
            }
            else
            {
                RemapFramesParser parser(fileP, &indices, sourceClip->GetVideoInfo().num_frames);
                try
                {
                    parser.parseReplaceSimple();
                }
                catch (RemapFramesParser::MalformedException&)
                {
                    envP->ThrowError("ReplaceFramesSimple: parse error "
                                     "(%s, line %u, column %u)",
                                     filenameP, parser.getLineNumber(), parser.getColumn());
                }
                catch (RemapFramesParser::OverflowException&)
                {
                    envP->ThrowError("ReplaceFramesSimple: integer overflow "
                                     "(%s, line %u, column %u)",
                                     filenameP, parser.getLineNumber(), parser.getColumn());
                }
                catch (RemapFramesParser::BadValueException& e)
                {
                    envP->ThrowError("ReplaceFramesSimple: value out of bounds: %d "
                                     "(%s, line %u)",
                                     e.val, filenameP, parser.getLineNumber());
                }
            }
        }

        if (mappingsP != NULL)
        {
            RemapFramesParser parser(mappingsP, &indices, sourceClip->GetVideoInfo().num_frames);
            try
            {
                parser.parseReplaceSimple();
            }
            catch (RemapFramesParser::MalformedException&)
            {
                envP->ThrowError("ReplaceFramesSimple: parse error in <mappings> string "
                                 "(line offset %u, column %u)",
                                 parser.getLineNumber(), parser.getColumn());
            }
            catch (RemapFramesParser::OverflowException&)
            {
                envP->ThrowError("ReplaceFramesSimple: integer overflow in <mappings> string "
                                 "(line offset %u, column %u)",
                                 parser.getLineNumber(), parser.getColumn());
            }
            catch (RemapFramesParser::BadValueException& e)
            {
                envP->ThrowError("ReplaceFramesSimple: value out of bounds in <mappings> string: %d "
                                 "(line offset %u)",
                                 e.val, parser.getLineNumber());
            }
        }
    }
    catch (std::bad_alloc&)
    {
        envP->ThrowError("ReplaceFramesSimple: insufficient memory");
    }
}


/** initAdvancedMode
  *
  *     Initializer for RemapFrames.
  *
  * PARAMETERS:
  *     IN filenameP - the name of the text file containing the frame
  *                      mappings;
  *                    may be NULL
  *     IN mappingsP - string containing additional frame mappings;
  *                    may be NULL
  *     IN/OUT envP  - pointer to the AviSynth scripting environment
  *
  * PRE:
  *     At least one of filenameP or mappingsP must not be NULL.
  */
void RemapFrames::initAdvancedMode(const char* filenameP, const char* mappingsP,
                                   IScriptEnvironment* envP)
{
    if (filenameP == NULL && mappingsP == NULL)
    {
        envP->ThrowError("RemapFrames: no filename nor mappings specified");
    }

    try
    {
        // Initialize the values in indices.
        // Each frame by default gets mapped to itself.
        indices.resize(vi.num_frames);
        for (int i = 0; i < vi.num_frames; i++)
        {
            indices[i].clipIndex = 0;
            indices[i].frame = i;
        }

        if (filenameP != NULL)
        {
            FILE* fileP = fopen(filenameP, "r");
            ON_BLOCK_EXIT(fclose, fileP);

            if (fileP == NULL)
            {
                envP->ThrowError("RemapFrames: error opening file \"%s\"", filenameP);
            }
            else
            {
                RemapFramesParser parser(fileP, &indices, sourceClip->GetVideoInfo().num_frames);
                try
                {
                    parser.parse();
                }
                catch (RemapFramesParser::MalformedException&)
                {
                    envP->ThrowError("RemapFrames: parse error "
                                     "(%s, line %u, column %u)",
                                     filenameP, parser.getLineNumber(), parser.getColumn());
                }
                catch (RemapFramesParser::OverflowException&)
                {
                    envP->ThrowError("RemapFrames: integer overflow "
                                     "(%s, line %u, column %u)",
                                     filenameP, parser.getLineNumber(), parser.getColumn());
                }
                catch (RemapFramesParser::BadValueException& e)
                {
                    envP->ThrowError("RemapFrames: value out of bounds: %d "
                                     "(%s, line %u)",
                                     e.val, filenameP, parser.getLineNumber());
                }
            }
        }

        if (mappingsP != NULL)
        {
            RemapFramesParser parser(mappingsP, &indices, sourceClip->GetVideoInfo().num_frames);
            try
            {
                parser.parse();
            }
            catch (RemapFramesParser::MalformedException&)
            {
                envP->ThrowError("RemapFrames: parse error in <mappings> string "
                                 "(line offset %u, column %u)",
                                 parser.getLineNumber(), parser.getColumn());
            }
            catch (RemapFramesParser::OverflowException&)
            {
                envP->ThrowError("RemapFrames: integer overflow in <mappings> string "
                                 "(line offset %u, column %u)",
                                 parser.getLineNumber(), parser.getColumn());
            }
            catch (RemapFramesParser::BadValueException& e)
            {
                envP->ThrowError("RemapFrames: value out of bounds in <mappings> string: %d "
                                 "(line offset %u)",
                                 e.val, parser.getLineNumber());
            }
        }
    }
    catch (std::bad_alloc&)
    {
        envP->ThrowError("RemapFrames: insufficient memory");
    }
}


/** RemapFrames constructor
  *
  * PARAMETERS:
  *     IN child_      - pointer to the base clip;
  *                      must have > 0 frames
  *     IN sourceClip_ - pointer to the source clip;
  *                      must have > 0 frames
  *     IN filenameP   - the name of the text file containing the frame
  *                        mappings;
  *                      may be NULL
  *     IN mappingsP   - string containing additional frame mappings;
  *                      may be NULL
  *     IN/OUT envP    - pointer to the AviSynth scripting environment
  */
RemapFrames::RemapFrames(PClip child_, PClip sourceClip_, mode_t mode,
                         const char* filenameP, const char* mappingsP,
                         IScriptEnvironment* envP)
: GenericVideoFilter(child_),
  sourceClip(sourceClip_),
  indices()
{
    assert(vi.num_frames > 0);
    assert(sourceClip->GetVideoInfo().num_frames > 0);

    switch (mode)
    {
        case MODE_SIMPLE:
            initSimpleMode(filenameP, mappingsP, envP);
            break;

        case MODE_REPLACE_SIMPLE:
            initReplaceSimpleMode(filenameP, mappingsP, envP);
            break;

        case MODE_ADVANCED:
            initAdvancedMode(filenameP, mappingsP, envP);
            break;

        default:
            assert(false);
    }
}


/** GetFrame
  *
  * PARAMETERS:
  *     n           - the index of the frame to retrieve
  *     IN/OUT envP - pointer to the AviSynth scripting environment
  *
  * RETURNS:
  *     the specified video frame
  */
PVideoFrame __stdcall RemapFrames::GetFrame(int n, IScriptEnvironment* envP)
{
    MapIndex& element = indices[n];
    return ((element.clipIndex == 0)
            ? child
            : sourceClip)->GetFrame(element.frame, envP);
}


/** GetParity
  *
  * PARAMETERS:
  *     n - the index of the frame
  *
  * RETURNS:
  *     the parity of the specified video frame
  */
bool __stdcall RemapFrames::GetParity(int n)
{
    MapIndex& element = indices[n];
    return ((element.clipIndex == 0)
            ? child
            : sourceClip)->GetParity(element.frame);
}


/** Create
  *
  *     Creates a new instance of this filter.
  *
  *     Callback passed to the scripting environment.
  *
  * PARAMETERS:
  *     args             - the array of arguments passed to this filter
  *                          from the script;
  *                        see filter documentation for details
  *     IN/OUT userDataP - (unused)
  *     IN/OUT envP      - pointer to the AviSynth scripting environment
  *
  * RETURNS:
  *     the new video clip
  */
AVSValue __cdecl RemapFrames::Create(AVSValue args, void* userDataP, IScriptEnvironment* envP)
{
    const PClip& clip = args[0].AsClip();
    const char* filenameP = args[1].AsString();
    const char* mappingsP = args[2].AsString();
    const PClip& sourceClip = args[3].Defined()
                              ? args[3].AsClip()
                              : clip;

    return (clip->GetVideoInfo().num_frames > 0 && sourceClip->GetVideoInfo().num_frames > 0)
           ? AVSValue(new RemapFrames(clip, sourceClip, MODE_ADVANCED, filenameP, mappingsP, envP))
           : args[0]; // nothing to do; return the base clip unaltered
}


AVSValue __cdecl RemapFrames::CreateSimple(AVSValue args, void* userDataP, IScriptEnvironment* envP)
{
    const PClip& clip = args[0].AsClip();
    const char* filenameP = args[1].AsString();
    const char* mappingsP = args[2].AsString();

    return (clip->GetVideoInfo().num_frames > 0)
           ? AVSValue(new RemapFrames(clip, clip, MODE_SIMPLE, filenameP, mappingsP, envP))
           : args[0]; // nothing to do; return the base clip unaltered
}


AVSValue __cdecl RemapFrames::CreateReplaceSimple(AVSValue args, void* userDataP, IScriptEnvironment* envP)
{
    const PClip& clip = args[0].AsClip();
    const PClip& sourceClip = args[1].AsClip();
    const char* filenameP = args[2].AsString();
    const char* mappingsP = args[3].AsString();

    return (clip->GetVideoInfo().num_frames > 0 && sourceClip->GetVideoInfo().num_frames > 0)
           ? AVSValue(new RemapFrames(clip, sourceClip, MODE_REPLACE_SIMPLE, filenameP, mappingsP, envP))
           : args[0]; // nothing to do; return the base clip unaltered
}


/** AvisynthPluginInit2
  *
  *     AviSynth 2.5x plug-in entry-point.
  *
  * PARAMETERS:
  *     IN/OUT envP - pointer to the AviSynth scripting environment
  *
  * RETURNS:
  *     the version string for this plug-in
  *
  * SIDE EFFECTS:
  *     adds functions to the AviSynth scripting environment
  */
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* envP)
{
    envP->AddFunction("RemapFrames", "c[filename]s[mappings]s[sourceClip]c", RemapFrames::Create, NULL);
    envP->AddFunction("RemapFramesSimple", "c[filename]s[mappings]s", RemapFrames::CreateSimple, NULL);
    envP->AddFunction("ReplaceFramesSimple", "cc[filename]s[mappings]s", RemapFrames::CreateReplaceSimple, NULL);
    return "RemapFrames v0.3.1 [" __DATE__ "]\nCopyright (c) 2005 James D. Lin";
}
